// tgMIDI class for doing MIDI I/O
//
// (c) 2002 by Tom Grandgent (tom@grandgent.com), started on 7/5/2002

#ifndef TGMIDI_DEFINED
#define TGMIDI_DEFINED

#include <windows.h>
#include <mmsystem.h>
#include <crtdbg.h>
#include <stdio.h>
#include <io.h>

#define HINYBBLE(a)		((a) >> 4)
#define LONYBBLE(a)		((a) & 15)
#define MAKEBYTE(l,h)	((l) | ((h) << 4))
#define RANDRANGE(a,b)	((a) + (rand() % ((b) - (a) + 1)))

// Display flags for drawKeyboard()
#define DK_NOTES		1				// Highlight currently-playing notes
#define DK_NOTENAMES	2				// Display names of highlighted notes

typedef struct midi_device_t {
	char *name;							// Driver-defined name for this device
	int deviceID;						// Device ID
	char enabled;						// Is this device currently enabled for playback?
	char input_device;					// Is this an input device?
	MIDIINCAPS incaps;					// Input device caps
	MIDIOUTCAPS outcaps;				// Output device caps

	midi_device_t *next;
} midi_device_t;

void CALLBACK MidiInProc(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2);

extern char *gmProgramNames[128];

class tgMIDI
{
private:
	// MIDI I/O handles
	HMIDIIN hin;
	HMIDIOUT hout;
	// MIDI device ID's of the chosen devices (initialized to -1, no device chosen)
	int inDeviceID, outDeviceID;
	// Note on/off array (for tracking)
	unsigned char noteVelocities[128];
	// Keyboard drawing parameters
	int kbInitialized;
	HWND kbHwnd;
	RECT kbRect;
	int kbLowNote, kbHighNote;
	// HWND to use as parent for error messages
	HWND hwndApp;
	// Critical section used to serialize access to MIDI out device (don't trust device drivers!)
	CRITICAL_SECTION csOut;

	int traverseKeyboard(unsigned char *noteArray, int displayFlags, int draw, int mx, int my);

public:
	midi_device_t *midi_devices;		// MIDI device list

	tgMIDI(void);
	~tgMIDI(void);
	void enumDevices(void);
	void openMidiIn(int newMidiInID);
	void openMidiOut(int newMidiOutID);
	void closeMidiIn(void);
	void closeMidiOut(void);
	void sendShortMessage(DWORD dwMsg);
	int getOutputDeviceID(void) { return outDeviceID; }
	int getInputDeviceID(void) { return inDeviceID; }

	void allNotesOffChannel(int channel, int abrupt);
	void allNotesOff(int abrupt);
	void noteOn(unsigned char note, unsigned char velocity, unsigned char channel);
	void noteOff(unsigned char note, unsigned char channel);
	void programChange(unsigned char program, unsigned char channel);
	void controlChange(unsigned char controller, unsigned char value, unsigned char channel);
	void pitchBend(short int value, unsigned char channel);
	unsigned char getNoteVelocity(unsigned char note);

	int sysexDump(FILE *fp);
	int sysexDump(unsigned char *buf, int len);

	char *noteNumToName(int noteNum);
	char *noteNumToNameWithoutOctave(int noteNum);
	int noteNameToNum(char *name);
	int isBlackKey(int noteNum);

	void setKeyboardParameters(HWND hwnd, RECT *r, int lowNote, int highNote);
	void drawKeyboard(unsigned char *noteArray, int displayFlags);
	int pointToNote(int mx, int my);

	void handleInputEvent(HMIDIIN hMidiIn, UINT wMsg, DWORD dwParam1, DWORD dwParam2);
};

#endif